#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use File::Basename;


###############################################################
##################### DOCUMENTATION ###########################

## Author: Yun Li
## Date: 
## Goal: 
my $location = $0;
my $title = "Dosages to scores Version 0.0.3";

print "$title\n";
print "\t\@: $location\n";
print "\tby: Yun Li(ylwtx\@umich.edu)\n\n";

## changelog
## V0.0.3
##       * Added $opts{tr} option
## V0.0.2
##       * implemented risky option
##
## Input Parameters
##

## Input Files
##

## Output Files

## Note:

################# END OF DOCUMENTATION ########################
###############################################################


my $MIN_NUM_OPTS = 3;

if($#ARGV<($MIN_NUM_OPTS*2 - 1))
{
	&usage();
	die "Failed to parse parameters\n\n";
}


my %opts = ();

	# Default Optional Options
$opts{o} = "dosages2score.out";
$opts{rareL} =  0.001; # 10 copies out of 10,000
$opts{rareU} = 0.05;# 5%;
$opts{risky} = 1;
$opts{tr} = 0.3;
my $ncontrols = 0;

Getopt::Long::GetOptions(\%opts,qw(
	csd=s
	ctd=s
	info=s
	rareL=f
	rareU=f
	risky=i
	tr=f
	o=s
)) || die "Failed to parse options\n\n";

my $TRAINING = $opts{tr}; # 30% used as training: to find the direction of minor alleles;

&printPar();

my $NUM_INFO_FIELDS = 7;
my $CASE_CODING = 2;
my $CONTROL_CODING = 1;
my $rareL = $opts{rareL};
my $rareU = $opts{rareU};

my @snps = ();
my @minoralleles = ();
my @weights = ();
my @alleleones = ();
my @alleletwos = ();
my @alleleonecounts = ();
my @adjmafs = (); # (num_minor_allele_among_controls + 1) / (2 + 2*n_controls)

my @riskyMarkers = ();

&main();


# begin sub

sub usage
{
	print "\n";
	print "Usage:\n\t";
	print "-csd\t Case dosages (Required) \n\t";
	print "-ctd\t Control dosages (Required) \n\t";
	print "-info\t Imputation info file (Required) \n\t";
	print "-rareL\t MAF lower bound for rare markers (Optional, default = 0.001) \n\t";
	print "-rareU\t MAF upper bound for rare markers (Optional, default = 0.05) \n\t";
	print "-risky\t Identify direction of effect (Optional, default = 1) \n\t";
	print "-tr\t Proportion used for training (Optional, default = 0.3) \n\t";
	print "-o\t Output (Optional) \n\t";
	print "\n";
}


sub printPar
{
	print "\n";
	print "Parameters in Effect:\n";
	print "\t Case dosages \n\t\t-csd '$opts{csd}'\n";
	print "\t Control dosages \n\t\t-ctd '$opts{ctd}'\n";
	print "\t Imputation info file \n\t\t-info '$opts{info}'\n";
	print "\t MAF lower bound for rare markers \n\t\t-rareL '$opts{rareL}'\n";
	print "\t MAF upper bound for rare markers \n\t\t-rareU '$opts{rareU}'\n";
	print "\t Identify direction of effect \n\t\t\-risky '$opts{risky}'\n";
	print "\t Proportion used for training \n\t\t\-tr '$opts{tr}'\n";
	print "\t Output \n\t\t-o '$opts{o}'\n";
	print "\n";
}





sub rm
{
	if(-e $_[0])
	{
		print "WARNING: $_[0] existed and removed!!\n\n";
		system "/bin/rm -rf $_[0]";
	}
}



sub bakrm
{
	if(-e $_[0])
	{
		my $tt = $_[0].".old";
		print "WARNING: $_[0] existed and moved to $tt !!\n\n";
		system "mv $_[0] $tt";
	}
}


sub swap
{
	my $tmp = $_[0];
	$_[0] = $_[1];
	$_[1] = $tmp;
}

sub main
{
  &loadInfo();
  if ($opts{risky} ne 0) {&findRiskyMarkers();}
  &calcControlMAF();
  &calcScores();
}


sub loadInfo
{
  open(IN, $opts{info}) || die "cannot open file $opts{info}\n\n";
  while(my $line = <IN>)
  {
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $NUM_INFO_FIELDS)
    {
      print "ERROR: expect $NUM_INFO_FIELDS fields from $opts{info}, the following line is invalid\n $line\n\n";
      exit;
    }
    push @snps, $linearr[0];
    push @alleleones, $linearr[1];
    push @alleletwos, $linearr[2];
  }
  close(IN);
  
  printf("Read %d SNPs from $opts{info}\n\n", scalar(@snps));
  
}

sub findRiskyMarkers
{
  # now use samples in training set to find the list of risky and protective markers
  
  # calculate case frequency
  my @casefreq1s = (); # assume no missing
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $casefreq1s[$i] = 0;
  }
  my $expectedFields = 2 + scalar(@snps);
  my $ntotalcases = 0;
  open(IN,$opts{csd}) || die "cannot open file $opts{csd}\n\n";
  while(my $line = <IN>)
  {
    $ntotalcases ++;
  }
  close(IN);
  my $trcases = $ntotalcases * $TRAINING;

  my $linenum = 0;
  open(IN,$opts{csd}) || die "cannot open file $opts{csd}\n\n";
  while(my $line = <IN>)
  {
    $linenum ++;
    if ($linenum > $trcases) {last;}
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $expectedFields)
    {
      die "ERROR: dosage file must have exactly $expectedFields fields while $linenum in $opts{csd} does not\n\n";
    }
    
    for (my $i = 0; $i <= $#snps; $i++)
    {
      $casefreq1s[$i] += $linearr[$i + 2];
    }
  }
  close(IN);
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $casefreq1s[$i] /= $linenum * 2;
  }
  print "Found marker frequencies on $linenum cases in training set\n";
    
  # calculate control frequency
  my @controlfreq1s = (); # assume no missing
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $controlfreq1s[$i] = 0;
  }
  
  my $ntotalcontrols = 0;

  open(IN,$opts{ctd}) || die "cannot open file $opts{ctd}\n\n";
  while(my $line = <IN>)
  {
    $ntotalcontrols ++;
  }
  my $ntrct = $ntotalcontrols * $TRAINING;

  $linenum = 0;
  open(IN,$opts{ctd}) || die "cannot open file $opts{ctd}\n\n";
  while(my $line = <IN>)
  {
    $linenum ++;
    if ($linenum > $ntrct) {last;}
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $expectedFields)
    {
      die "ERROR: dosage file must have exactly $expectedFields fields while $linenum in $opts{ctd} does not\n\n";
    }
    
    for (my $i = 0; $i <= $#snps; $i++)
    {
      $controlfreq1s[$i] += $linearr[$i + 2];
    }
  }
  close(IN);

  my $ncontrols = $linenum;
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $controlfreq1s[$i] /= $ncontrols * 2;
  }
  print "Found marker frequencies on $ncontrols controls\n";
  
  # find markers that show allele frequency difference
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $riskyMarkers[$i] = 0;
    my $controlmaf = $controlfreq1s[$i];
    my $caseFreqForControlMinorAllele = $casefreq1s[$i];
    if ($controlfreq1s[$i] > 0.5)
    {
      $controlmaf = 1 - $controlfreq1s[$i];
      $caseFreqForControlMinorAllele = 1 - $casefreq1s[$i];
    }
    
    my $cutoff = 1.28 * sqrt($controlmaf*(1-$controlmaf)/2/$ncontrols);
    my $allelediff = $caseFreqForControlMinorAllele - $controlmaf;
    if ($allelediff > 0 && $allelediff > $cutoff)
    {
      $riskyMarkers[$i] = 1;
    }
    elsif ($allelediff < 0 && abs($allelediff) > $cutoff)
    {
      $riskyMarkers[$i] = -1;
    }
  }
  
}

sub calcControlMAF
{
  open(IN, $opts{ctd}) || die "cannot open file $opts{ctd}\n\n";
  for (my $i = 0; $i <= $#snps; $i++)
  {
    $alleleonecounts[$i] = 0;
  }
  
  my $expectedFields = 2 + scalar(@snps);
  # count # controls
  my $linenum = 0;
  while(my $line = <IN>)
  {
    $linenum ++;
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $expectedFields)
    {
      die "ERROR: dosage file must have exactly $expectedFields fields while $linenum in $opts{ctd} does not\n\n";
    }
    
    for (my $i = 0; $i <= $#snps; $i++)
    {
      $alleleonecounts[$i] += $linearr[$i + 2];
    }
  }
  close(IN);
  $ncontrols = $linenum;
  print "Read $ncontrols controls from $opts{ctd}\n\n";
  
  for(my $i = 0; $i <= $#snps; $i ++)
  {
    if ($alleleonecounts[$i] > $ncontrols) 
    {
      $minoralleles[$i] = $alleletwos[$i];
      $adjmafs[$i] = (2*$ncontrols - $alleleonecounts[$i] + 1) / (2 + 2*$ncontrols);
    }
    else 
    {
      $minoralleles[$i] = $alleleones[$i];
      $adjmafs[$i] = ($alleleonecounts[$i] + 1) / (2 + 2*$ncontrols);
    }
    
    if ($adjmafs[$i] >= $opts{rareL} && $adjmafs[$i] <= $opts{rareU})
    {
      $weights[$i] = 1.0 / (sqrt($ncontrols * $adjmafs[$i] * (1.0 - $adjmafs[$i])));
    }
    else
    {
      $weights[$i] = 0.0;
    }
  }
}

sub calcScores
{
  my $expectedFields = 2 + scalar(@snps);
  
  &bakrm($opts{o});
  open(OUT, ">>$opts{o}") || die "cannot create file $opts{o}\n\n";
  print OUT "id affection score\n";

  #deal with controls
  Open($opts{ctd},\*IN) || die "cannot open file $opts{ctd}\n\n";
  my $linenum = 0;
  while(my $line = <IN>)
  {
    $linenum ++;
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $expectedFields)
    {
      die "ERROR: dosage file must have exactly $expectedFields fields while $linenum in $opts{ctd} does not\n\n";
    }
    
    my $score = 0.0;
    my $riskyvalue = 1;
    for (my $i = 0; $i <= $#snps; $i ++)
    {
      if ($opts{risky} ne 0) {$riskyvalue = $riskyMarkers[$i];}
      my $index = $i + 2;
      if ($minoralleles[$i] eq $alleletwos[$i])
      {
        $score += (2.0 - $linearr[$index]) * $weights[$i] * $riskyvalue;
      }
      else
      {
        $score += $linearr[$index] * $weights[$i] * $riskyvalue;
      }
    }
    print OUT "$linearr[0] $CONTROL_CODING $score\n";
  }
  close(IN);
  #end dealing with controls
  
  # deal with cases
  Open($opts{csd},\*IN) || die "cannot open file $opts{csd}\n\n";
  $linenum = 0;
  while(my $line = <IN>)
  {
    $linenum ++;
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $expectedFields)
    {
      die "ERROR: dosage file must have exactly $expectedFields fields while $linenum in $opts{csd} does not\n\n";
    }
    
    my $score = 0.0;
    my $riskyvalue = 1;
    for (my $i = 0; $i <= $#snps; $i ++)
    {
      my $index = $i + 2;
      if ($opts{risky} ne 0) {$riskyvalue = $riskyMarkers[$i];}
      if ($minoralleles[$i] eq $alleletwos[$i])
      {
        $score += (2.0 - $linearr[$index]) * $weights[$i] * $riskyvalue;
      }
      else
      {
        $score += $linearr[$index] * $weights[$i] * $riskyvalue;
      }
    }
    print OUT "$linearr[0] $CASE_CODING $score\n";
  }
  close(IN);
  # end dealing with cases
  
  close(OUT);
}


#==================================================================
# Subroutine:
#   Open($filename, $handle)
#
# Usage:
#   Open('myfile.txt', \*IN);
#   Open($pedfile, \*PED);
#
#   Then refernce the handles <IN> and <PED> as usual.
#
# Description:
#   Open the file 'filename'. The file can be a normal file
#   or one compressed with gzip or bzip2.
#   Function does not return if the open fails.
#
# Arguments:
#   $filename - path to file to open
#   $handle - GLOB name of handle to use. Include '\*'
#
#==================================================================
sub Open {
    my ($filename, $handle) = @_;

    # Handle gzipped data
    if ($filename =~ /\.gz$/) {
        my $cmd = "gunzip -c $filename";
        open($handle, "$cmd |") ||          
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 1;
    }
    # Handle bzip2 data
    if ($filename =~ /\.bz2$/) {
        my $cmd = "bunzip2 -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 2;     
    }              
    # Handle normal file
    open($handle, $filename) ||
        die "Unable to open file '$filename': $!\n";
    return 3;
}                  




 
