#!/bin/csh -f

## Input parameters:
##                  (1) csh: case haplotypes
##                  (2) cth: control haplotypes
##                  (3) col: haplotype column 
##                  (4) out: output prefix
## Notes:
##       (1) 30% of cases and 30% of controls will be used for training
##       (2) # cases can differ from # controls
##       (3) input haplotype file must be uncompressed (.gz file will NOT be accepted)
##       (4) Programs called:
##                          ** mergetables
##                          ** RHG.rph.r

##############################
## step 0: check parameters ##
##############################
set par = $1
set csh = `echo "$par" | cut -f1 -d ',' ` #case haplotypes
set cth = `echo "$par" | cut -f2 -d ',' ` #control haplotypes
set col = `echo "$par" | cut -f3 -d ',' ` #haplotype column
set out = `echo "$par" | cut -f4 -d ',' ` #output prefix
set tr = `echo "$par" | cut -f5 -d ',' ` #% training

if ($tr == "") then
  echo "WARNING: % training not specified. Using default 30%"
  $tr = 0.3
endif

if ($csh == "" || $cth == "" || $col == "" || $out == "") then
  echo "Usage: haps2riskyHap.training.csh cases.hap,controls.hap,hap_column,prefix"
  echo ""
  exit
endif

###########################
## step 1: initial setup ##
###########################
  ## find out number of case/control haplotypes ##
@ ncsh = `wc -l < $csh`
@ ncth = `wc -l < $cth`
  ## calculate # of case/control haplotypes in training ##
@ trcsh = `echo "$ncsh * $tr" | bc | cut -f1 -d '.' `
@ ntrcs = $trcsh / 2
@ trcth = `echo "$ncth * $tr" | bc | cut -f1 -d '.' `
@ ntrct = $trcth / 2
  ## calculate # of diploid cases/controls in testing ##
@ nttcs = ($ncsh - $trcsh) / 2
@ nttct = ($ncth - $trcth) / 2


####################################################################################
## step 2: calculate haplotype counts for cases and controls in the training set ##
####################################################################################
  ## cases ##
echo "hap cnt" > $out.cases.training.cnts
head -$trcsh $csh | awk -v c=$col '{print $c}' | sort | uniq -c | awk '{print $2,$1}' >> $out.cases.training.cnts

  ## controls ##
echo "hap cnt" > $out.controls.training.cnts
head -$trcth $cth | awk -v c=$col '{print $c}' | sort | uniq -c | awk '{print $2,$1}' >> $out.controls.training.cnts

  ## set of unique haplotypes in training set ##
echo "hap" > $out.training.uniq_hap
cat $out.cases.training.cnts $out.controls.training.cnts | grep -v hap | awk '{print $1}' | sort | uniq >> $out.training.uniq_hap

#########################################################################
## step 3: merge case and control haplotype counts in the training set ##
##         compare to get the set of risk rare haplotypes 
#########################################################################
mergetables --outer-join $out.training.uniq_hap $out.controls.training.cnts $out.cases.training.cnts | awk '{print "A"$1,$3,$5}' | sed -e 's/cnt.3/case_counts/' -e 's/cnt/control_counts/' -e 's/MISS/0/g' > $out.training.contrast
rm -f $out.training.uniq_hap
rm -f $out.controls.training.cnts
rm -f $out.cases.training.cnts

##########################################################################
## step 4: identify the set of risk rare haplotypes in the training set ##
##########################################################################
if ($ncsh == $ncth) then
  cat ~/bin/RHG.rph.r | R --vanilla --args $out.training.contrast $out.training.rp.hap $ntrcs > $out.training.contrast.rlog
else
  cat ~/bin/RHG.rph.imbalanced.r | R --vanilla --args $out.training.contrast $out.training.rp.hap $ntrcs $ntrct > $out.training.contrast.rlog
endif
rm -f $out.training.contrast.rlog
rm -f $out.training.contrast
