#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use File::Basename;


###############################################################
##################### DOCUMENTATION ###########################

## Author: Yun Li
## Date: 
## Goal: 
my $location = $0;
my $title = "haplotypes to Rare-Haplotype-Weighed scores Version 0.0.5";

print "$title\n";
print "\t\@: $location\n";
print "\tby: Yun Li(ylwtx\@med.unc.edu)\n\n";

## changelog
##          0.0.5
##		           (1) add -risky option to feed in list of risky AND protective haplotypes (second column +/-)
##		               default: all
##		               if provided: other haplotypes will have weight zero

##          0.0.4
##		           (1) add -risky option to feed in list of riskly haplotypes
##		       default: all
##		       if provided: other haplotypes will have weight zero

##          0.0.3
##               (1) allow user to specify MAF cutoffs 


## Input Parameters
##

## Input Files
##

## Output Files

## Note:
##     (1) require exact three fields of .hap

################# END OF DOCUMENTATION ########################
###############################################################

my $MIN_NUM_OPTS = 2;

if($#ARGV<($MIN_NUM_OPTS*2 - 1))
{
	&usage();
	die "Failed to parse parameters\n\n";
}


my %opts = ();

	# Default Optional Options
$opts{o} = "haps2score.out";
$opts{col} = 3;
$opts{rareL} =  0.001; # 10 copies out of 10,000
$opts{rareU} = 0.05;# 5%;
$opts{risky} = 0;

Getopt::Long::GetOptions(\%opts,qw(
	csh=s
	cth=s
	rareL=f
	rareU=f
	col=i
	risky=s
	o=s
)) || die "Failed to parse options\n\n";


&printPar();

my $NUM_HAP_FIELDS = $opts{col};
my $CASE_CODING = 2;
my $CONTROL_CODING = 1;
my $rareL = $opts{rareL};
my $rareU = $opts{rareU};

my %riskyHaps = ();
my %controlHapFreq = (); # corrected control haplotype frequencies: 
                         # key = haplotype 
                         # value = q = (#minor_alleles_among_controls + 1) / (2+2*#controls)
my %weights = ();        # key = haplotype 
                         # value = 1/sqrt(#controls*q*(1-q))
&main();


# begin sub

sub usage
{
	print "\n";
	print "Usage:\n\t";
	print "-csh\t Case Haplotypes (Required) \n\t";
	print "-cth\t Control Haplotypes (Required) \n\t";
	print "-rareL\t MAF lower bound for rare haplotypes (Optional, default = 0.001) \n\t";
	print "-rareU\t MAF upper bound for rare haplotypes (Optional, default = 0.05) \n\t";
	print "-col\t Haplotype Field (Optional, default=3) \n\t";
	print "-risky\t List of risky haplotypes (Optional, all haplotypes will be considered by default) \n\t";
	print "-o\t Output (Optional) \n\t";
	print "\n";
}


sub printPar
{
	print "\n";
	print "Parameters in Effect:\n";
	print "\t Case Haplotypes \n\t\t-csh '$opts{csh}'\n";
	print "\t Control Haplotypes \n\t\t-cth '$opts{cth}'\n";
	print "\t MAF lower bound for rare haplotypes \n\t\t-rareL '$opts{rareL}'\n";
	print "\t MAF upper bound for rare haplotypes \n\t\t-rareU '$opts{rareU}'\n";
	print "\t Haplotype Field \n\t\t\-col '$opts{col}'\n";
	print "\t List of risky haplotypes \n\t\t\-risky '$opts{risky}'\n";
	print "\t Output \n\t\t-o '$opts{o}'\n";
	print "\n";
}





sub rm
{
	if(-e $_[0])
	{
		print "WARNING: $_[0] existed and removed!!\n\n";
		system "/bin/rm -rf $_[0]";
	}
}



sub bakrm
{
	if(-e $_[0])
	{
		my $tt = $_[0].".old";
		print "WARNING: $_[0] existed and moved to $tt !!\n\n";
		system "mv $_[0] $tt";
	}
}


sub swap
{
	my $tmp = $_[0];
	$_[0] = $_[1];
	$_[1] = $tmp;
}

sub main
{
  if ($opts{risky} ne 0) {&loadRiskyHap();}
  &calcControlHapFreq();
  &calcScores();
}

sub loadRiskyHap
{
  open(IN,$opts{risky}) || die "Cannot open file $opts{risky}\n\n";
  my $linenum = 0;
  while(my $line = <IN>)
  {
    $linenum ++;
    chomp($line);
    my @linearr = (split(/\s+/, $line));
    if (scalar(@linearr) != 2)
    {
      print "ERROR: each line in $opts{risky} needs to contain exactly two fields. Line $linenum is invalid\n\n";
      exit;
    }
    if ($linearr[1] ne "+" && $linearr[1] ne "-")
    {
      print "ERROR: invalid flag in line $linenum of opts{risky}\n\n";
      exit;
    }
    if (exists $riskyHaps{$linearr[0]})
    {
      print "ERROR: $linearr[0] duplicated as a risky haplotype\n\n";
      exit;
    }
    if ($linearr[1] eq "+") {$riskyHaps{$linearr[0]} = 1;}
    if ($linearr[1] eq "-") {$riskyHaps{$linearr[0]} = -1;}
  }
  close(IN);
}

sub calcControlHapFreq
{
  open(IN, $opts{cth}) || die "cannot open file $opts{cth}\n\n";
  
  # count haplotypes
  my $linenum = 0;
  while(my $line = <IN>)
  {
    $linenum ++;
    chomp($line);
    my @linearr = (split(/\s+/,$line));
    if (scalar(@linearr) != $NUM_HAP_FIELDS)
    {
      die "ERROR: haplotype file must have exactly $NUM_HAP_FIELDS fields while $linenum in $opts{cth} does not\n\n";
    }
    
    if (! exists $controlHapFreq{$linearr[$#linearr]})
    {
      $controlHapFreq{$linearr[$#linearr]} = 1;
    }
    else
    {
      $controlHapFreq{$linearr[$#linearr]} ++;
    }
  }
  close(IN);
  
  #while ((my $key, my $value) = each(%controlHapFreq)) {print $key." ".$value."\n";}
  #print "\n";
  
  # convert counts to corrected frequencies
  my $numRareHaps = 0;
  while ((my $key, my $value) = each(%controlHapFreq))
  {
    $value = ($value + 1) / ($linenum + 2);
    $controlHapFreq{$key} = $value;
    if ($value >= $rareL && $value <= $rareU)
    {
      $numRareHaps ++;
    }
  }
  
  print "Read $linenum control haplotypes with " . keys(%controlHapFreq) . " unique haplotypes ($numRareHaps are rare)\n";
 
  #while ((my $key, my $value) = each(%controlHapFreq)) {print $key." ".$value."\n";}
  
  while ((my $key, my $value) = each(%controlHapFreq))
  {
    if ( ($opts{risky} eq 0) || exists $riskyHaps{$key})
    {
      $weights{$key} = 1/sqrt($linenum*$controlHapFreq{$key}*(1-$controlHapFreq{$key})/2);
      if (exists $riskyHaps{$key})
      {
        $weights{$key} *= $riskyHaps{$key};
      }
      #$weights{$key} = 1;
      #print $key." ".$weights{$key}."\n";
    }
    else
    {
      $weights{$key} = 0;
    }
  }
  
  # assign weights for novel haplotypes
  my $baseControlHapFreq = 1 / ($linenum + 2);
  $weights{"UNSEEN"} = 1/sqrt($linenum*$baseControlHapFreq*(1-$baseControlHapFreq)/2);
  #$weights{"UNSEEN"} = 1;
  
 
}

sub calcScores
{
  &bakrm($opts{o});
  open(OUT, ">>$opts{o}") || die "cannot create file $opts{o}\n\n";
  print OUT "id affection score\n";

  #deal with controls
  open(IN, $opts{cth}) || die "cannot open file $opts{cth}\n\n";
  my $linenum = 0;
  while(my $hapone = <IN>)
  {
    $linenum ++;
    chomp($hapone);
    my @haponearr = (split(/\s+/,$hapone));
    if (scalar(@haponearr) != $NUM_HAP_FIELDS)
    {
      die "ERROR: haplotype file must have exactly $NUM_HAP_FIELDS fields while $linenum in $opts{cth} does not\n\n";
    }
    
    my $haptwo = <IN>;
    $linenum ++;
    chomp($haptwo);
    my @haptwoarr = (split(/\s+/,$haptwo));
    if (scalar(@haptwoarr) != $NUM_HAP_FIELDS)
    {
      die "ERROR: haplotype file must have exactly $NUM_HAP_FIELDS fields while $linenum in $opts{cth} does not\n\n";
    }
    
    if ($opts{col} > 1 && ($haponearr[0] ne $haptwoarr[0]))
    {
      printf("ERROR: line %d and %d of $opts{cth} are not for the same individual\n\n", $linenum - 1, $linenum);
      die;
    }
    
    my $score = 0;
    if (exists $weights{$haponearr[$#haponearr]}) 
    {
      if ($controlHapFreq{$haponearr[$#haponearr]} >= $rareL && $controlHapFreq{$haponearr[$#haponearr]} <= $rareU)
      {
        $score += $weights{$haponearr[$#haponearr]};
      }
    }
    if (exists $weights{$haptwoarr[$#haptwoarr]}) 
    {
      if ($controlHapFreq{$haptwoarr[$#haptwoarr]} >= $rareL && $controlHapFreq{$haptwoarr[$#haptwoarr]} <= $rareU)
      {
        $score += $weights{$haptwoarr[$#haptwoarr]};
      }
    }
    print OUT "$haponearr[0] $CONTROL_CODING $score\n";
    
  }
  close(IN);
  #end dealing with controls
  
  # deal with cases
  open(IN, $opts{csh}) || die "cannot open file $opts{csh}\n\n";
  $linenum = 0;
  while(my $hapone = <IN>)
  {
    $linenum ++;
    chomp($hapone);
    my @haponearr = (split(/\s+/,$hapone));
    if (scalar(@haponearr) != $NUM_HAP_FIELDS)
    {
      die "ERROR: haplotype file must have exactly $NUM_HAP_FIELDS fields while $linenum in $opts{csh} does not\n\n";
    }
    
    my $haptwo = <IN>;
    $linenum ++;
    chomp($haptwo);
    my @haptwoarr = (split(/\s+/,$haptwo));
    if (scalar(@haptwoarr) != $NUM_HAP_FIELDS)
    {
      die "ERROR: haplotype file must have exactly $NUM_HAP_FIELDS fields while $linenum in $opts{csh} does not\n\n";
    }
    
    if ($opts{col} > 1 && ($haponearr[0] ne $haptwoarr[0]))
    {
      printf("ERROR: line %d and %d of $opts{csh} are not for the same individual\n\n", $linenum - 1, $linenum);
      die;
    }
    
    my $score = 0;
    if (exists $weights{$haponearr[$#haponearr]}) 
    {
      if ($controlHapFreq{$haponearr[$#haponearr]} >= $rareL && $controlHapFreq{$haponearr[$#haponearr]} <= $rareU)
      {
        $score += $weights{$haponearr[$#haponearr]};
      }
    }# hap1 seen among controls
    else
    {
      if ($opts{risky} eq 0 || exists $riskyHaps{$haponearr[$#haponearr]}) 
      {
        my $weight = $weights{"UNSEEN"};
        if (exists $riskyHaps{$haponearr[$#haponearr]}) {$weight *= $riskyHaps{$haponearr[$#haponearr]};}
        $score += $weight;
      }
    }# hap1 not seen in controls
    
    if (exists $weights{$haptwoarr[$#haptwoarr]}) 
    {
      if ($controlHapFreq{$haptwoarr[$#haptwoarr]} >= $rareL && $controlHapFreq{$haptwoarr[$#haptwoarr]} <= $rareU)
      {
        $score += $weights{$haptwoarr[$#haptwoarr]};
      }
    }# hap2 seen among controls
    else
    {
      if ($opts{risky} eq 0 || exists $riskyHaps{$haptwoarr[$#haptwoarr]}) 
      {
        my $weight = $weights{"UNSEEN"};
        if (exists $riskyHaps{$haptwoarr[$#haptwoarr]}) {$weight *= $riskyHaps{$haptwoarr[$#haptwoarr]};}
        $score += $weight;
      }
    }# hap2 not seen in controls
    
    print OUT "$haponearr[0] $CASE_CODING $score\n";
    
  }
  close(IN);
  # end dealing with cases
  
  close(OUT);
}




 
